#!/usr/local/bin/python3

# S.D.G.

# Name: mglycanstring.py
# Description: mglycanstring is a simulator of enzyme activities potentially involved in HMO biosynthesis
# Author: Andrew G. McDonald
# ORCID iD: https://orcid.org/0000-0003-2727-176X
# Date: 4 May 2022

# Usage:
# python3 mglycanstring.py <iterations> [options]

# Options:
# 	-a	Use refined enzyme rules that match additional substrate-patterns
# 	-ln	Limit the number of GlcNAc residues to n per glycan (unused)
# 	-plist	Choose enzymes to knock out, as a comma-separated list of assignments of the form name=val, 
# 		where name is the short enzyme name and val is 0 (knockout) or 1 (knock-in).
# 		Example: -pb4GalT=0,ST6GlcNAc=0
# 	-sstring
#		Set the initial substrate to value of string (default is free lactose "[L4G]"). Example: -s"[Y3L4G]"

# Symbol key:
# Letter  IUPAC      Definition                                      Default anomer
# f       Fuc        L-Fucose                                        &alpha;
# G       Glc        D-Glucose                                       &beta;
# K       Kdn        2-Keto-3-deoxy-D-glycero-D-galacto-nononic acid &alpha;
# L       Gal        D-Galactose                                     &beta;
# M       Man        D-Mannose                                       &alpha;
# N       Neu5Gc     N-Glycolylneuraminic acid                       &alpha;
# S       Neu5Ac     N-Acetylneuraminic acid (sialic acid)           &alpha;
# u       IdoA       L-Iduronic acid                                 &alpha;
# V       GalNAc     N-Acetyl-D-galactosamine                        &alpha;
# X       Xyl        D-Xylose                                        &beta;
# Y       GlcNAc     N-Acetyl-D-glucosamine                          &beta;
# Z       GlcA       D-Glucuronic acid                               &beta;
# s       -HSO3      Sulfate                                         n/a
# a,b     alpha,beta anomeric type                                   n/a


import sys, re

# define the enzymes used by the simulator as a tuple

enzymes = (
"beta4GalT",
"alpha4FucT",
"alpha2FucT",
"beta3GnT",
"alpha3FucT",
"beta3GalT",
"ST6Gal",
"ST3Gal",
"ST6GlcNAc",
"cIGnT",
"dIGnT"
)


# set default parameter values

p = {}
for e in enzymes:
	p[e] = 1.0

limit = 0
# "advanced" mode activates further enzyme-substrate recognition patterns:
#  use command-line switch '-a' to enable
advanced = ""
halt = 0
# halt becomes true (1) if limit > 0 and c > limit
# set initial substrate (free lactose)
glycan = "[L4G]"

# parse the command line

if len(sys.argv)>1:
	maxiter = int(sys.argv[1])
	arglist = sys.argv[2:]
	for item in arglist:
		if item == '-a':
			advanced = '-a'
			print ("Using advanced model")
		if item.startswith('-l'):
			limit = int(item.replace('-l','',1))
			print ("Limiting to level", limit)
		if item.startswith('-p'):
			kolist = item.replace('-p','',1).split(",")
			for ep in kolist:
				pair = ep.split("=")
				p[pair[0]] = int(pair[1])
		if item.startswith('-s'):
			glycan = item.replace('-s','',1)

else:
	print ("No args supplied")
	maxiter = 1



# define the remaining hashes (dictionaries)

newp = {}
news = newp
pool = {}

newg = 0
newp[glycan] = glycan

if glycan not in pool:
	pool[glycan] = glycan
	newg += 1
else:
	newg = 0
n = {}
n[glycan] = len(pool)
product = ""
iteration = 0

# function definitions
def react( string, action ):
	global advanced
	if action == "beta4GalT":
		string = re.sub(r'\[(\[s6\])?Y([36])', r"[L4\1Y\2", string, 1)
	elif action == "alpha4FucT":
		string = re.sub(r'\[L3(\[s6\])?Y', r"[L3[f4]\1Y", string, 1)
		string = re.sub(r'\[L4G\]', r"[L4[f3]G]", string, 1)
	elif action == "alpha2FucT":
		string = re.sub(r'\[L([34])(\[f[34]\])?(\[s6\])?(\]V|[GY])', r"[[f2]L\1\2\3\4", string, 1)
	elif action == "beta3GnT":
		if advanced :
			string = re.sub(r'\[(\[s[6]\])?L4(\[f[34]\])?(\[s6\])?([GY])', r"[Y3\1L4\2\3\4", string, 1)
		else :
			string = re.sub(r'\[(\[s[6]\])?L4(\[s6\])?([GY])', r"[Y3\1L4\2\3", string, 1)
	elif action == "alpha3FucT":
		string = re.sub(r'\[L4(\[s6\])?([GY])(.)', r"[L4[f3]\1\2\3", string, 1)
	elif action == "beta3GalT":
		string = re.sub(r'\[(\[s6\])?Y([36])', r"[L3\1Y\2", string, 1)
	elif action == "ST6Gal":
		string = re.sub(r'\[L4(\[f[34]\])?(\[s6\])?(G|Y[36])', r"[S6L4\1\2\3", string, 1)
	elif action == "ST3Gal":
		string = re.sub(r'\[L([34])(\[f[34]\])?(\[s6\])?([GY])', r"[S3L\1\2\3\4", string, 1)
	elif action == "ST6GlcNAc":
		string = re.sub(r'\[([KNS]3)?L3Y3', r"[\1L3[S6]Y3", string, 1)
	elif action == "cIGnT":
		string = re.sub(r'\[(\[s[36]\])?L4(\[s6\])?Y3L(4|3\]V)', r"[[Y6][\1L4\2Y3]L\3", string, 1)
	elif action == "dIGnT":
		string = re.sub(r'\[(\[s6\])?Y3L(4|3\]V)', r"[[Y6][\1Y3]L\2", string, 1)
	return string;

def chunk1( substrate, enzyme ):
	# for substrates with a single pattern-match
	global newg
	retstr = ""
	product = react(substrate,enzyme);
	if product != '' and substrate != product :
		print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
		if product not in pool :
			newg += 1
			newp[product] = product
		pool[product] = substrate
		print (product + " (" + str(len(pool)) + ")");
	return

def chunk1a( substrate, enzyme ):
	# for substrates of a GnT (GlcNAc-transferase) with a single pattern-match
	global newg, newp, halt, limit
	retstr = ""
	product = react(substrate,enzyme);
	c = product.count('Y')
	if limit and limit > c :
		halt = 1
	if halt < 1 and product != '' and substrate != product :
		print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
		if product not in pool :
			newg += 1
			newp[product] = product
		pool[product] = substrate
		n[product] = len(pool)
		print (product + " (" + str(len(pool)) + ")");
	return

def chunk2( regex, substrate, enzyme ):
	# for substrates having multiple pattern-matches
	global newg, newp, halt, limit
	matches = re.split(regex, substrate)
	l = len(matches)-1
	if l >= 3 and 1 == 1 :
		nextsubstrate = ""
		for i in range(l, 0, -1):
			nextsubstrate = matches[i] + nextsubstrate
			prefix = ""
			for j in range(0, i):
				prefix = prefix + matches[j]
			if i % 2 == 1 and i < l :
				nextproduct = prefix + react(nextsubstrate, enzyme)
				if nextproduct != '' and nextproduct != nextsubstrate :
					print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
					if nextproduct not in pool :
						newg += 1
						newp[nextproduct] = nextproduct
					pool[nextproduct] = prefix + nextsubstrate
					print (nextproduct + " (" + str(len(pool)) + ")");
	else :
		chunk1(substrate,enzyme)
	return

def chunk2a( regex, substrate, enzyme ):
	# for substrates of a GnT (GlcNAc-transferase) having multiple pattern-matches
	global newg, newp, halt, limit
	matches = re.split(regex, substrate)
	l = len(matches)-1
	if l >= 3 and 1 == 1 :
		nextsubstrate = ""
		for i in range(l, 0, -1):
			nextsubstrate = matches[i] + nextsubstrate
			prefix = ""
			for j in range(0, i):
				prefix = prefix + matches[j]
			if i % 2 == 1 and i < l :
				nextproduct = prefix + react(nextsubstrate, enzyme)
				c = nextproduct.count('Y')
				if limit and limit > c :
					halt = 1
				if halt < 1 and nextproduct != '' and nextsubstrate != nextproduct :
					print (str(iteration) + ": " + substrate + " -- " + enzyme + " --> ", end="")
					if nextproduct not in pool :
						newg += 1
						newp[nextproduct] = nextproduct
					pool[nextproduct] = prefix + nextsubstrate
					print (nextproduct + " (" + str(len(pool)) + ")");
	else :
		chunk1a(substrate,enzyme)
	return

# start of main loop

while newg > 0 and iteration < maxiter :
	iteration += 1
	newg = 0

	news = newp	# the new products of one iteration become the substrates of the next
	newp = {}

	# cycle through the list of new substrates
	for substrate in news:

		# 1. Extension: beta-1,4-galactosyltransferase
		enzyme = "beta4GalT"
		if p[enzyme] > 0:
                        if re.search('\[\[s6\]Y[36]', substrate) :
                                chunk2('(\[\[s6\]Y[36])',substrate,enzyme)
                        if re.search('\[Y[36]', substrate) :
                                chunk2('(\[Y[36])',substrate,enzyme)

		# 2. Decoration: alpha-1,4-fucosyltransferase
		enzyme = "alpha4FucT"
		if p[enzyme] > 0:
                        if re.search('\[L3\[s6\]Y', substrate) :
                                chunk2('(\[L3\[s6\]Y)',substrate,enzyme)
                        if re.search('\[L3Y', substrate) :
                                chunk2('(\[L3Y)',substrate,enzyme)
                        # 3-FL formation
                        if re.search(r'\[L4G\]', substrate) :
                                chunk1(substrate,enzyme)

		# 3. Decoration: alpha-1,2-fucosyltransferase
		enzyme = "alpha2FucT"
		if p[enzyme] > 0:
                        if re.search('\[L[34]\[f[34]\]\[s6\][GY]', substrate) :
                                chunk2('(\[L[34]\[f[34]\]\[s6\][GY])',substrate,enzyme)
                        if re.search('\[L[34]\[f[34]\][GY]', substrate) :
                                chunk2('(\[L[34]\[f[34]\][GY])',substrate,enzyme)
                        if re.search('\[L[34]\[s6\][GY]', substrate) :
                                chunk2('(\[L[34]\[s6\][GY])',substrate,enzyme)
                        if re.search('\[L[34][GY]', substrate) :
                                chunk2('(\[L[34][GY])',substrate,enzyme)
                        # core 1/2 Gal support (redundant)
                        if re.search('\[L3\]V', substrate) :
                                chunk1(substrate,enzyme)

		# 4. Extension: lactosylceramide alpha-2,3-sialyltransferase
		enzyme = "beta3GnT"
		if p[enzyme] > 0:
                        # use 'and not...' to add further boolean restrictions on enzyme action
                        if re.search('\[\[s[6]\]L4\[s6\][GY]', substrate): #and not re.search('Y6\]\[.+Y3\]L4', substrate):
                                chunk2a('(\[\[s[6]\]L4\[s6\][GY])',substrate,enzyme)
                        if re.search('\[L4\[s6\][GY]', substrate): #and not re.search('Y6\]\[.+Y3\]L4', substrate):
                                chunk2a('(\[L4\[s6\][GY])',substrate,enzyme)
                        if re.search('\[\[s[6]\]L4[GY]', substrate): #and not re.search('Y6\]\[.+Y3\]L4', substrate):
                                chunk2a('(\[\[s[6]\]L4[GY])',substrate,enzyme)
                        if re.search('\[L4[GY]', substrate): #and not re.search('Y6\]\[.+Y3\]L4', substrate):
                                chunk2a('(\[L4[GY])',substrate,enzyme)
                        if advanced :
                                if re.search('\[\[s[6]\]L4\[f3\]\[s6\][GY]', substrate): #and not re.search('Y6\]\[.+Y3\]L4', substrate):
                                        chunk2a('(\[\[s[6]\]L4\[f3\]\[s6\][GY])',substrate,enzyme)
                                if re.search('\[L4\[f3\]\[s6\][GY]', substrate): #and not re.search('Y6\]\[.+Y3\]L4', substrate):
                                        chunk2a('(\[L4\[f3\]\[s6\][GY])',substrate,enzyme)
                                if re.search('\[\[s[6]\]L4\[f3\][GY]', substrate): #and not re.search('Y6\]\[.+Y3\]L4', substrate):
                                        chunk2a('(\[\[s[6]\]L4\[f3\][GY])',substrate,enzyme)
                                if re.search('\[L4\[f3\][GY]', substrate): #and not re.search('Y6\]\[.+Y3\]L4', substrate):
                                        chunk2a('(\[L4\[f3\][GY])',substrate,enzyme)

		# 5. Decoration: alpha-N-acetylneuraminate alpha-2,8-sialyltransferase
		enzyme = "alpha3FucT"
		if p[enzyme] > 0:
                        if re.search('\[L4\[s6\]Y.', substrate) :
                                chunk2('(\[L4\[s6\]Y.)',substrate,enzyme)
                        if re.search('\[L4Y.', substrate) :
                                chunk2('(\[L4Y.)',substrate,enzyme)

		# 6. Extension: (alpha-2,8-N-acetylneuraminate alpha-2,8-sialyltransferase)
		enzyme = "beta3GalT"
		if p[enzyme] > 0:
                        if re.search('\[\[s6\]Y[36]', substrate) :
                                chunk2('(\[\[s6\]Y[36])',substrate,enzyme)
                        if re.search('\[Y[36]', substrate) :
                                chunk2('(\[Y[36])',substrate,enzyme)

		# 7. Termination: (N-acetylneuraminyl)-galactosylglucosylceramide N-acetylgalactosaminyltransferase
		enzyme = "ST6Gal"
		if p[enzyme] > 0:
                        if re.search('\[L4\[f[34]\]\[s6\][GY]', substrate) :
                                chunk2('(\[L4\[f[34]\]\[s6\][GY])',substrate,enzyme)
                        if re.search('\[L4\[f[34]\][GY]', substrate) :
                                chunk2('(\[L4\[f[34]\][GY])',substrate,enzyme)
                        if re.search('\[L4\[s6\][GY]', substrate) :
                                chunk2('(\[L4\[s6\][GY])',substrate,enzyme)
                        if re.search('\[L4[GY]', substrate) :
                                chunk2('(\[L4[GY])',substrate,enzyme)

		# 8. Termination: N-acetyllactosaminide alpha-2,3-sialyltransferase
		enzyme = "ST3Gal"
		if p[enzyme] > 0:
                        if re.search('\[L[34]\[f[34]\]\[s6\][GY]', substrate) :
                                chunk2('(\[L[34]\[f[34]\]\[s6\][GY])',substrate,enzyme)
                        if re.search('\[L[34]\[f[34]\][GY]', substrate) :
                                chunk2('(\[L[34]\[f[34]\][GY])',substrate,enzyme)
                        if re.search('\[L[34]\[s6\][GY]', substrate) :
                                chunk2('(\[L[34]\[s6\][GY])',substrate,enzyme)
                        if re.search('\[L[34][GY]', substrate) :
                                chunk2('(\[L[34][GY])',substrate,enzyme)

		# 9. Decoration: (beta-1,3-galactosyl-ceramide alpha-2,3-sialyltransferase)
		enzyme = "ST6GlcNAc"
		if p[enzyme] > 0:
			chunk1(substrate,enzyme)

		# 10. Branching: (alpha1,3-Sia-beta1,3-Gal-beta1,3-GalNAc alpha-2,6-sialyltransferase)
		enzyme = "cIGnT"
		if p[enzyme] > 0 and not re.search(r'L4\[f3\]G\]', substrate):
                        if re.search('\[(\[s[36]\])?L4(\[s6\])?Y3L4', substrate) :
                                if re.search('\[\[s[36]\]L4\[s6\]Y3L', substrate) :
                                        chunk2a('(\[\[s[36]\]L4\[s6\]Y3L)',substrate,enzyme)
                                if re.search('\[L4\[s6\]Y3L4', substrate) :
                                        chunk2a('(\[L4\[s6\]Y3L4)',substrate,enzyme)
                                if re.search('\[\[s[36]\]L4Y3L4', substrate) :
                                        chunk2a('(\[\[s[36]\]L4Y3L4)',substrate,enzyme)
                                if re.search('\[L4Y3L4', substrate) :
                                        chunk2a('(\[L4Y3L4)',substrate,enzyme)
                        if re.search('\[(\[s[36]\])?L4(\[s6\])?Y3L3\]V', substrate) :
                                chunk1a(substrate,enzyme)

		# 11. Branching: (alpha1,3-Sia-beta1,3-Gal-beta1,3-GalNAc alpha-2,6-sialyltransferase)
		enzyme = "dIGnT"
		if p[enzyme] > 0 and not re.search(r'L4\[f3\]G\]', substrate):
                        if re.search('\[\[s6\]Y3L4', substrate) :
                                chunk2a('(\[\[s6\]Y3L4)',substrate,enzyme)
                        if re.search('\[Y3L4', substrate) :
                                chunk2a('(\[Y3L4)',substrate,enzyme)

	if newg == 0 :
		print ("Nothing to be done!")
	else :
		print ("New structures: " + str(newg))

# print a list of all generated species

print ("\nHMO pool:")

for key in sorted(pool):
	print (key)

